@@ -1,25 +1,58 @@
-#' Nonparametric Variable Importance Estimates and Inference using Cross-fitting
+#' Nonparametric Intrinsic Variable Importance Estimates and Inference using Cross-fitting
 #'
-#' Compute estimates and confidence intervals using cross-fitting for nonparametric variable importance based on the population-level contrast between the oracle predictiveness using the feature(s) of interest versus not.
+#' Compute estimates and confidence intervals using cross-fitting for 
+#' nonparametric intrinsic variable importance based on the 
+#' population-level contrast between the oracle predictiveness using the 
+#' feature(s) of interest versus not.
 #'
 #' @param Y the outcome.
 #' @param X the covariates.
-#' @param f1 the predicted values on validation data from a flexible estimation technique regressing Y on X in the training data; a list of length V, where each object is a set of predictions on the validation data.
-#' @param f2 the predicted values on validation data from a flexible estimation technique regressing the fitted values in \code{f1} on X withholding the columns in \code{indx}; a list of length V, where each object is a set of predictions on the validation data.
-#' @param indx the indices of the covariate(s) to calculate variable importance for; defaults to 1.
+#' @param f1 the predicted values on validation data from a flexible estimation 
+#'   technique regressing Y on X in the training data; a list of length V, where
+#'   each object is a set of predictions on the validation data.
+#' @param f2 the predicted values on validation data from a flexible estimation 
+#'   technique regressing the fitted values in \code{f1} on X withholding the 
+#'   columns in \code{indx}; a list of length V, where each object is a set of 
+#'   predictions on the validation data.
+#' @param indx the indices of the covariate(s) to calculate variable importance 
+#'   for; defaults to 1.
 #' @param V the number of folds for cross-fitting, defaults to 10.
-#' @param folds the folds to use, if f1 and f2 are supplied. A list of length two; the first element provides the outer folds (for hypothesis testing), while the second element is a list providing the inner folds (for cross-fitting).
-#' @param stratified if run_regression = TRUE, then should the generated folds be stratified based on the outcome (helps to ensure class balance across cross-fitting folds)
+#' @param folds the folds to use, if f1 and f2 are supplied. A list of length 
+#'   two; the first element provides the outer folds (for hypothesis testing), 
+#'   while the second element is a list providing the inner folds 
+#'   (for cross-fitting).
+#' @param stratified if run_regression = TRUE, then should the generated folds 
+#'   be stratified based on the outcome (helps to ensure class balance across 
+#'   cross-fitting folds)
 #' @param type the type of parameter (e.g., ANOVA-based is \code{"anova"}).
-#' @param run_regression if outcome Y and covariates X are passed to \code{cv_vim}, and \code{run_regression} is \code{TRUE}, then Super Learner will be used; otherwise, variable importance will be computed using the inputted fitted values.
-#' @param SL.library a character vector of learners to pass to \code{SuperLearner}, if \code{f1} and \code{f2} are Y and X, respectively. Defaults to \code{SL.glmnet}, \code{SL.xgboost}, and \code{SL.mean}.
-#' @param alpha the level to compute the confidence interval at. Defaults to 0.05, corresponding to a 95\% confidence interval.
-#' @param delta the value of the \eqn{\delta}-null (i.e., testing if importance < \eqn{\delta}); defaults to 0.
-#' @param scale should CIs be computed on original ("identity") or logit ("logit") scale?
-#' @param na.rm should we remove NA's in the outcome and fitted values in computation? (defaults to \code{FALSE})
-#' @param C the indicator of coarsening (1 denotes observed, 0 denotes unobserved).
-#' @param Z either (i) NULL (the default, in which case the argument \code{C} above must be all ones), or (ii) a character vector specifying the variable(s) among Y and X that are thought to play a role in the coarsening mechanism.
-#' @param ipc_weights weights for the computed influence curve (i.e., inverse probability weights for coarsened-at-random settings). Assumed to be already inverted (i.e., ipc_weights = 1 / [estimated probability weights]).
+#' @param run_regression if outcome Y and covariates X are passed to 
+#'   \code{cv_vim}, and \code{run_regression} is \code{TRUE}, then Super Learner 
+#'   will be used; otherwise, variable importance will be computed using the 
+#'   inputted fitted values.
+#' @param SL.library a character vector of learners to pass to 
+#'   \code{SuperLearner}, if \code{f1} and \code{f2} are Y and X, respectively. 
+#'   Defaults to \code{SL.glmnet}, \code{SL.xgboost}, and \code{SL.mean}.
+#' @param alpha the level to compute the confidence interval at. Defaults to 
+#'   0.05, corresponding to a 95\% confidence interval.
+#' @param delta the value of the \eqn{\delta}-null (i.e., testing if 
+#'   importance < \eqn{\delta}); defaults to 0.
+#' @param scale should CIs be computed on original ("identity") or 
+#'   logit ("logit") scale?
+#' @param na.rm should we remove NA's in the outcome and fitted values in 
+#'   computation? (defaults to \code{FALSE})
+#' @param C the indicator of coarsening (1 denotes observed, 0 denotes 
+#'   unobserved).
+#' @param Z either (i) NULL (the default, in which case the argument 
+#'   \code{C} above must be all ones), or (ii) a character vector specifying 
+#'   the variable(s) among Y and X that are thought to play a role in the 
+#'   coarsening mechanism.
+#' @param ipc_weights weights for the computed influence curve (i.e., inverse 
+#'   probability weights for coarsened-at-random settings). Assumed to be 
+#'   already inverted (i.e., ipc_weights = 1 / [estimated probability weights]).
+#' @param ipc_est_type the type of procedure used for coarsened-at-random 
+#'   settings; options are "ipw" (for inverse probability weighting) or 
+#'   "aipw" (for augmented inverse probability weighting). 
+#'   Only used if \code{C} is not all equal to 1.
 #' @param ... other arguments to the estimation tool, see "See also".
 #'
 #' @return An object of class \code{vim}. See Details for more information.
@@ -129,12 +162,21 @@
 #'
 #' @seealso \code{\link[SuperLearner]{SuperLearner}} for specific usage of the \code{SuperLearner} function and package.
 #' @export
-cv_vim <- function(Y = NULL, X = NULL, f1 = NULL, f2 = NULL, indx = 1, V = length(unique(folds)), folds = NULL, stratified = FALSE, type = "r_squared", run_regression = TRUE, SL.library = c("SL.glmnet", "SL.xgboost", "SL.mean"), alpha = 0.05, delta = 0, scale = "identity", na.rm = FALSE, C = rep(1, length(Y)), Z = NULL, ipc_weights = rep(1, length(Y)), ...) {
+cv_vim <- function(Y = NULL, X = NULL, f1 = NULL, f2 = NULL, indx = 1, 
+                   V = length(unique(folds)), folds = NULL, 
+                   stratified = FALSE, type = "r_squared", 
+                   run_regression = TRUE, 
+                   SL.library = c("SL.glmnet", "SL.xgboost", "SL.mean"), 
+                   alpha = 0.05, delta = 0, scale = "identity", 
+                   na.rm = FALSE, C = rep(1, length(Y)), Z = NULL, 
+                   ipc_weights = rep(1, length(Y)), 
+                   ipc_est_type = "aipw", ...) {
     # check to see if f1 and f2 are missing
     # if the data is missing, stop and throw an error
     check_inputs(Y, X, f1, f2, indx)
 
-    # check to see if Y is a matrix or data.frame; if not, make it one (just for ease of reading)
+    # check to see if Y is a matrix or data.frame; if not, make it one 
+    # (just for ease of reading)
     if(is.null(dim(Y))) Y <- as.matrix(Y)
 
     # set up internal data -- based on complete cases only
@@ -169,9 +211,15 @@ cv_vim <- function(Y = NULL, X = NULL, f1 = NULL, f2 = NULL, indx = 1, V = lengt
     if (run_regression) {
         X_cc <- subset(X, C == 1, drop = FALSE)
         # set up the cross-fitting
-        outer_folds <- .make_folds(Y, V = 2, stratified = stratified)
-        inner_folds_1 <- .make_folds(Y[outer_folds == 1, , drop = FALSE], V = V, stratified = stratified)
-        inner_folds_2 <- .make_folds(Y[outer_folds == 2, , drop = FALSE], V = V, stratified = stratified)
+        outer_folds <- .make_folds(Y, V = 2, C = C, stratified = stratified)
+        inner_folds_1 <- .make_folds(
+            Y[outer_folds == 1, , drop = FALSE], V = V, 
+            C = C[outer_folds == 1], stratified = stratified
+        )
+        inner_folds_2 <- .make_folds(
+            Y[outer_folds == 2, , drop = FALSE], V = V, 
+            C = C[outer_folds == 2], stratified = stratified
+        )
         outer_folds_cc <- outer_folds[C == 1]
         inner_folds_1_cc <- inner_folds_1[C[outer_folds == 1] == 1]
         inner_folds_2_cc <- inner_folds_2[C[outer_folds == 2] == 1]
@@ -181,39 +229,83 @@ cv_vim <- function(Y = NULL, X = NULL, f1 = NULL, f2 = NULL, indx = 1, V = lengt
         fhat_red <- list()
         for (v in 1:V) {
             # fit super learner
-            fit <- SuperLearner::SuperLearner(Y = Y_cc[(outer_folds_cc == 1), , drop = FALSE][inner_folds_1_cc != v, , drop = FALSE], X = X_cc[(outer_folds_cc == 1), , drop = FALSE][inner_folds_1_cc != v, , drop = FALSE], SL.library = SL.library, obsWeights = weights_cc[(outer_folds_cc == 1)][inner_folds_1_cc != v], ...)
-            fitted_v <- SuperLearner::predict.SuperLearner(fit, onlySL = TRUE)$pred
+            fit <- SuperLearner::SuperLearner(
+                Y = Y_cc[(outer_folds_cc == 1), , 
+                         drop = FALSE][inner_folds_1_cc != v, , drop = FALSE], 
+                X = X_cc[(outer_folds_cc == 1), , 
+                         drop = FALSE][inner_folds_1_cc != v, , drop = FALSE], 
+                SL.library = SL.library, 
+                obsWeights = weights_cc[(outer_folds_cc == 1)][inner_folds_1_cc != v], 
+                ...
+            )
+            fitted_v <- SuperLearner::predict.SuperLearner(
+                fit, onlySL = TRUE
+            )$pred
             # get predictions on the validation fold
-            fhat_ful[[v]] <- SuperLearner::predict.SuperLearner(fit, newdata = X_cc[(outer_folds_cc == 1), , drop = FALSE][inner_folds_1_cc == v, , drop = FALSE], onlySL = TRUE)$pred
+            fhat_ful[[v]] <- SuperLearner::predict.SuperLearner(
+                fit, 
+                newdata = X_cc[(outer_folds_cc == 1), , 
+                               drop = FALSE][inner_folds_1_cc == v, , drop = FALSE], 
+                onlySL = TRUE
+            )$pred
             # fit the super learner on the reduced covariates:
             # if the reduced set of covariates is empty, return the mean
             # otherwise, if type is r_squared or anova, always use gaussian; if first regression was mean, use Y instead
             if (ncol(X_cc[(outer_folds_cc == 2), , drop = FALSE][, -indx, drop = FALSE]) == 0) {
-                fhat_red[[v]] <- mean(Y_cc[(outer_folds_cc == 2), , drop = FALSE][inner_folds_2_cc != v, , drop = FALSE])
+                fhat_red[[v]] <- mean(
+                    Y_cc[(outer_folds_cc == 2), , 
+                         drop = FALSE][inner_folds_2_cc != v, , drop = FALSE]
+                )
             } else {
                 arg_lst <- list(...)
                 if (length(unique(fitted_v)) == 1) {
-                    arg_lst$Y <- Y_cc[(outer_folds_cc == 2), , drop = FALSE][inner_folds_2_cc != v, , drop = FALSE]
+                    arg_lst$Y <- Y_cc[(outer_folds_cc == 2), , 
+                                      drop = FALSE][inner_folds_2_cc != v, , 
+                                                    drop = FALSE]
                 } else if (type == "r_squared" || type == "anova") {
                     arg_lst$family <- stats::gaussian()
-                    fit_2 <- SuperLearner::SuperLearner(Y = Y_cc[(outer_folds_cc == 2), , drop = FALSE][inner_folds_2_cc != v, , drop = FALSE], X = X_cc[(outer_folds_cc == 2), , drop = FALSE][inner_folds_2_cc != v, , drop = FALSE], SL.library = SL.library, ...)
-                    arg_lst$Y <- SuperLearner::predict.SuperLearner(fit_2, onlySL = TRUE)$pred
+                    if (any(grepl("cvControl", names(arg_lst)))) {
+                        arg_lst$cvControl$stratifyCV <- FALSE
+                    }
+                    fit_2 <- SuperLearner::SuperLearner(
+                        Y = Y_cc[(outer_folds_cc == 2), , 
+                                 drop = FALSE][inner_folds_2_cc != v, , drop = FALSE], 
+                        X = X_cc[(outer_folds_cc == 2), , 
+                                 drop = FALSE][inner_folds_2_cc != v, , drop = FALSE], 
+                        SL.library = SL.library, ...
+                    )
+                    arg_lst$Y <- SuperLearner::predict.SuperLearner(
+                        fit_2, onlySL = TRUE
+                    )$pred
                 } else {
-                    arg_lst$Y <- Y_cc[(outer_folds_cc == 2), , drop = FALSE][inner_folds_2_cc != v, , drop = FALSE]
+                    arg_lst$Y <- Y_cc[(outer_folds_cc == 2), , 
+                                      drop = FALSE][inner_folds_2_cc != v, , 
+                                                    drop = FALSE]
                     # get the family
                     if (is.character(arg_lst$family))
-                        arg_lst$family <- get(arg_lst$family, mode = "function", envir = parent.frame())
+                        arg_lst$family <- get(arg_lst$family, mode = "function", 
+                                              envir = parent.frame())
                 }
-                arg_lst$X <- X_cc[(outer_folds_cc == 2), , drop = FALSE][inner_folds_2_cc != v, -indx, drop = FALSE]
+                arg_lst$X <- X_cc[(outer_folds_cc == 2), , 
+                                  drop = FALSE][inner_folds_2_cc != v, -indx, 
+                                                drop = FALSE]
                 arg_lst$SL.library <- SL.library
                 arg_lst$obsWeights <- weights_cc[(outer_folds_cc == 2)][inner_folds_2_cc != v]
                 red <- do.call(SuperLearner::SuperLearner, arg_lst)
                 # get predictions on the validation fold
-                fhat_red[[v]] <- SuperLearner::predict.SuperLearner(red, newdata = X_cc[(outer_folds_cc == 2), , drop = FALSE][inner_folds_2_cc == v, -indx, drop = FALSE], onlySL = TRUE)$pred
+                fhat_red[[v]] <- SuperLearner::predict.SuperLearner(
+                    red, 
+                    newdata = X_cc[(outer_folds_cc == 2), , 
+                                   drop = FALSE][inner_folds_2_cc == v, -indx, 
+                                                 drop = FALSE], onlySL = TRUE
+                )$pred
             }
         }
         full <- reduced <- NA
-        folds <- list(outer_folds = outer_folds, inner_folds = list(inner_folds_1, inner_folds_2))
+        folds <- list(
+            outer_folds = outer_folds, 
+            inner_folds = list(inner_folds_1, inner_folds_2)
+        )
 
     } else { # otherwise they are fitted values
         # check to make sure that the fitted values, folds are what we expect
@@ -231,9 +323,35 @@ cv_vim <- function(Y = NULL, X = NULL, f1 = NULL, f2 = NULL, indx = 1, V = lengt
 
         full <- reduced <- NA
     }
+    arg_lst <- list(...)
+    if (!is.null(names(arg_lst)) && any(grepl("cvControl", names(arg_lst)))) {
+        arg_lst$cvControl$stratifyCV <- FALSE
+    }
     # calculate the estimators, EIFs
     if (full_type == "anova" || full_type == "regression") {
-        est_lst <- sapply(1:V, function(v) measure_anova(fhat_ful[[v]], fhat_red[[v]], Y_cc[outer_folds_cc == 1, , drop = FALSE][inner_folds_1_cc == v], C[outer_folds == 1][inner_folds_1 == v], Z = Z_in[outer_folds == 1, , drop = FALSE][inner_folds_1 == v, , drop = FALSE], ipc_weights[outer_folds == 1][inner_folds_1 == v], ipc_fit_type = "SL", na.rm = na.rm, SL.library = SL.library,  ...), simplify = FALSE)
+        est_lst <- sapply(
+            1:V, 
+            function(v) 
+                do.call(
+                    measure_anova,
+                    args = c(
+                        list(full = fhat_ful[[v]], 
+                             reduced = fhat_red[[v]], 
+                             y = Y_cc[outer_folds_cc == 1, , 
+                                  drop = FALSE][inner_folds_1_cc == v], 
+                             C = C[outer_folds == 1][inner_folds_1 == v], 
+                             Z = Z_in[outer_folds == 1, , 
+                                      drop = FALSE][inner_folds_1 == v, , 
+                                                    drop = FALSE], 
+                             ipc_weights = ipc_weights[outer_folds == 1][inner_folds_1 == v], 
+                             ipc_fit_type = "SL", na.rm = na.rm, 
+                             ipc_est_type = ipc_est_type, scale = scale,
+                             SL.library = SL.library),
+                        arg_lst
+                    )
+                ), 
+            simplify = FALSE
+        )
         point_ests <- unlist(lapply(est_lst, function(x) x$point_est))
         naives <- unlist(lapply(est_lst, function(x) x$naive))
         est <- mean(point_ests)
@@ -250,16 +368,52 @@ cv_vim <- function(Y = NULL, X = NULL, f1 = NULL, f2 = NULL, indx = 1, V = lengt
         se_full <- rep(NA, V)
         se_redu <- rep(NA, V)
     } else {
-        est_lst_full <- est_predictiveness_cv(fitted_values = fhat_ful, y = Y_cc[outer_folds_cc == 1, , drop = FALSE], folds = inner_folds_1_cc, type = full_type, C = C[outer_folds == 1], Z = Z_in[outer_folds == 1, , drop = FALSE], folds_Z = inner_folds_1, ipc_weights = ipc_weights[outer_folds == 1], ipc_fit_type = "SL", scale = scale, na.rm = na.rm, SL.library = SL.library, ...)
-        est_lst_redu <- est_predictiveness_cv(fitted_values = fhat_red, y = Y_cc[outer_folds_cc == 2, , drop = FALSE], folds = inner_folds_2_cc, type = full_type, C = C[outer_folds == 2], Z = Z_in[outer_folds == 2, , drop = FALSE], folds_Z = inner_folds_2, ipc_weights = ipc_weights[outer_folds == 2], ipc_fit_type = "SL", scale = scale, na.rm = na.rm, SL.library = SL.library, ...)
+        est_lst_full <- do.call(
+            est_predictiveness_cv,
+            args = c(
+                list(fitted_values = fhat_ful, 
+                     y = Y_cc[outer_folds_cc == 1, , drop = FALSE], 
+                     full_y = Y_cc,
+                     folds = inner_folds_1_cc, type = full_type, 
+                     C = C[outer_folds == 1], 
+                     Z = Z_in[outer_folds == 1, , drop = FALSE], 
+                     folds_Z = inner_folds_1, 
+                     ipc_weights = ipc_weights[outer_folds == 1], 
+                     ipc_fit_type = "SL", scale = scale, 
+                     ipc_est_type = ipc_est_type, na.rm = na.rm, 
+                     SL.library = SL.library),
+                arg_lst
+            )
+        )
+        est_lst_redu <- do.call(
+            est_predictiveness_cv,
+            args = c(
+                list(fitted_values = fhat_red, 
+                     y = Y_cc[outer_folds_cc == 2, , drop = FALSE], 
+                     full_y = Y_cc,
+                     folds = inner_folds_2_cc, type = full_type, 
+                     C = C[outer_folds == 2], 
+                     Z = Z_in[outer_folds == 2, , drop = FALSE], 
+                     folds_Z = inner_folds_2, 
+                     ipc_weights = ipc_weights[outer_folds == 2], 
+                     ipc_fit_type = "SL", scale = scale, 
+                     ipc_est_type = ipc_est_type, na.rm = na.rm, 
+                     SL.library = SL.library),
+                arg_lst
+            )
+        )
         predictiveness_full <- est_lst_full$point_est
         predictiveness_redu <- est_lst_redu$point_est
         all_eifs_full <- est_lst_full$all_eifs
         eif_full <- est_lst_full$eif
         all_eifs_redu <- est_lst_redu$all_eifs
         eif_redu <- est_lst_redu$eif
-        se_full <- vimp_se(predictiveness_full, eif_full, na.rm = na.rm)
-        se_redu <- vimp_se(predictiveness_redu, eif_redu, na.rm = na.rm)
+        se_full <- vimp_se(list(est = predictiveness_full, 
+                                eif = eif_full,
+                                all_eifs = all_eifs_full), na.rm = na.rm)
+        se_redu <- vimp_se(list(est = predictiveness_redu, 
+                                eif = eif_redu,
+                                all_eifs = all_eifs_redu), na.rm = na.rm)
         est <- est_lst_full$point_est - est_lst_redu$point_est
         naive <- NA
         len_full <- length(eif_full)
@@ -272,21 +426,14 @@ cv_vim <- function(Y = NULL, X = NULL, f1 = NULL, f2 = NULL, indx = 1, V = lengt
             tmp_eif_full <- eif_full
             tmp_eif_redu <- eif_redu
         }
-        nrow_full <- nrow(all_eifs_full)
-        nrow_redu <- nrow(all_eifs_redu)
-        if (nrow_full != nrow_redu) {
-            max_nrow <- max(nrow_full, nrow_redu)
-            tmp_eifs_full <- rbind(all_eifs_full, matrix(NA, nrow = max_nrow - nrow_full, ncol = ncol(all_eifs_full)))
-            tmp_eifs_redu <- rbind(all_eifs_redu, matrix(NA, nrow = max_nrow - nrow_redu, ncol = ncol(all_eifs_redu)))
-        } else {
-            tmp_eifs_full <- all_eifs_full
-            tmp_eifs_redu <- all_eifs_redu
-        }
-        eif <- tmp_eif_full - tmp_eif_redu
-        all_eifs <- tmp_eifs_full - tmp_eifs_redu
+        eif <- tmp_eif_full - tmp_eif_redu 
+        all_eifs <- sapply(1:V, function(v) {
+            all_eifs_full[[v]] - all_eifs_redu[[v]]
+        }, simplify = FALSE)
     }
     # compute the standard error
-    se <- vimp_se(est, eif, na.rm = na.rm)
+    se <- vimp_se(list(est = est, eif = eif, all_eifs = all_eifs), 
+                  na.rm = na.rm)
 
     # if est < 0, set to zero and print warning
     if (est < 0 && !is.na(est)) {
@@ -298,19 +445,31 @@ cv_vim <- function(Y = NULL, X = NULL, f1 = NULL, f2 = NULL, indx = 1, V = lengt
 
     # calculate the confidence interval
     ci <- vimp_ci(est, se, scale = scale, level = 1 - alpha)
-    predictiveness_ci_full <- vimp_ci(predictiveness_full, se = se_full, scale = scale, level = 1 - alpha)
-    predictiveness_ci_redu <- vimp_ci(predictiveness_redu, se = se_redu, scale = scale, level = 1 - alpha)
+    predictiveness_ci_full <- vimp_ci(
+        predictiveness_full, se = se_full, scale = scale, level = 1 - alpha
+    )
+    predictiveness_ci_redu <- vimp_ci(
+        predictiveness_redu, se = se_redu, scale = scale, level = 1 - alpha
+    )
 
     # perform a hypothesis test against the null of zero importance
     if (full_type == "anova" || full_type == "regression") {
         hyp_test <- list(test = NA, p_value = NA, test_statistics = NA)
     } else {
-        hyp_test <- vimp_hypothesis_test(predictiveness_full = predictiveness_full, predictiveness_reduced = predictiveness_redu, se_full = se_full, se_reduced = se_redu, delta = delta, alpha = alpha)
+        hyp_test <- vimp_hypothesis_test(
+            predictiveness_full = predictiveness_full, 
+            predictiveness_reduced = predictiveness_redu, 
+            se_full = se_full, se_reduced = se_redu, 
+            delta = delta, alpha = alpha
+        )
     }
 
     # create the output and return it (as a tibble)
     chr_indx <- paste(as.character(indx), collapse = ",")
-    mat <- tibble::tibble(s = chr_indx, est = est, se = se[1], cil = ci[1], ciu = ci[2], test = hyp_test$test, p_value = hyp_test$p_value)
+    mat <- tibble::tibble(
+        s = chr_indx, est = est, se = se[1], cil = ci[1], ciu = ci[2], 
+        test = hyp_test$test, p_value = hyp_test$p_value
+    )
     output <- list(s = chr_indx,
                  SL.library = SL.library,
                  full_fit = fhat_ful, red_fit = fhat_red,
